"""
This sample illustrates how to get a range map from a blaze camera and a color image
from a 2D GigEVision (GEV) color camera using the Python Harvester module.

A Basler blaze-101 camera and a Basler 2D GEV color camera are required for this sample.
"""

import os
import platform

## This is used for reshaping the image buffers.
import numpy as np

## This is used for visualization and debayering.
import cv2

## Use of Harvester to access the camera.
## For more information regarding Harvester, visit the github page:
## https://github.com/genicam/harvesters
from harvesters.core import Harvester


BAYER_FORMATS = {"BayerGR8" : cv2.COLOR_BayerGR2RGB, \
        "BayerRG8" : cv2.COLOR_BayerRG2RGB, \
        "BayerBG8" : cv2.COLOR_BayerBG2RGB, \
        "BayerGB8" : cv2.COLOR_BayerGB2RGB}


def find_producer(name):
    ## Helper for the GenTL producers from the environment path.
    paths = os.environ['GENICAM_GENTL64_PATH'].split(os.pathsep)

    for path in paths:
        path += os.path.sep + name
        if os.path.exists(path):
            return path
    return ""

class ColorAndDepth:
    def __init__(self):
        ## Create Harvester instances.
        self.h = Harvester()

        ## Location of the Basler blaze GenTL producer.
        if platform.system() == "Windows":
            path_to_blaze_cti = os.environ['BASLER_BLAZE_BIN64'] + "\\ProducerBlaze.cti"
            path_to_gev_cti = find_producer("ProducerGEV.cti")
        elif platform.system() == "Linux":
            path_to_blaze_cti = "/opt/Basler_blaze/lib/gentlproducer/gtl/ProducerBlaze.cti"
            path_to_gev_cti = find_producer("ProducerGEV.cti")
        else:
            print(f"{platform.system()} is not supported")
            assert False

        ## Add producer to Harvester.
        assert os.path.exists(path_to_blaze_cti)
        assert os.path.exists(path_to_gev_cti)

        self.h.add_file(path_to_blaze_cti)
        self.h.add_file(path_to_gev_cti)

        ## Update device list.
        self.h.update()

        ## Print device list.
        print(self.h.device_info_list)


    def setup_blaze(self):
        ## Connect to first blaze-101 camera.
        dev_info = next((d for d in self.h.device_info_list if d.model == 'blaze-101'), None)
        if dev_info is not None:
            self.ia_blaze = self.h.create_image_acquirer(model=dev_info.model)
        else:
            print("No blaze camera found.")
            exit()

        ## Set the component selector.
        ## Range information can be sent either as a 16-bit grey value image or
        ## as 3D coordinates (point cloud).
        ## For this sample, we want to acquire the range map.
        ## Note: To change the format of an image component, the Component Selector parameter
        ## must first be set to the component
        ## you want to configure.
        ## To use 16-bit integer depth information, choose "Coord3D_C16" instead
        ## of "Coord3D_ABC32f".
        self.ia_blaze.remote_device.node_map.ComponentSelector.value = "Range"
        self.ia_blaze.remote_device.node_map.ComponentEnable.value = True
        self.ia_blaze.remote_device.node_map.PixelFormat.value = "Coord3D_C16"

        self.ia_blaze.remote_device.node_map.ComponentSelector.value = "Intensity"
        self.ia_blaze.remote_device.node_map.ComponentEnable.value = True
        self.ia_blaze.remote_device.node_map.PixelFormat.value = "Mono16"

        self.ia_blaze.remote_device.node_map.ComponentSelector.value = "Confidence"
        self.ia_blaze.remote_device.node_map.ComponentEnable.value = True
        self.ia_blaze.remote_device.node_map.PixelFormat.value = "Confidence8"

        ## Configure the camera for software triggering.
        self.ia_blaze.remote_device.node_map.TriggerMode.value = "On"
        self.ia_blaze.remote_device.node_map.TriggerSource.value = "Software"

        ## Start image acquisition.
        self.ia_blaze.start_acquisition()

    def setup_2Dcamera(self):
        ## Connect to the first available 2D camera, ignore blaze cameras which will
        ## be enumerated too.
        dev_info = next((d for d in self.h.device_info_list if 'blaze' not in d.model), None)
        if dev_info is not None:
            self.ia_gev = self.h.create_image_acquirer(serial_number=dev_info.serial_number)
        else:
            print("No 2D camera found.")
            exit()

        bayer_pattern = next((pf for pf in self.ia_gev.remote_device.node_map.PixelFormat.symbolics
                              if pf in BAYER_FORMATS), None)
        if bayer_pattern is None:
            print("The camera does not provide Bayer-pattern encoded 8 bit color images")
            exit()

        ## Configure the camera for software triggering.
        ## Each software trigger will start the acquisition of one single frame.
        self.ia_gev.remote_device.node_map.AcquisitionMode.value = "Continuous"
        self.ia_gev.remote_device.node_map.TriggerSelector.value = "FrameStart"
        self.ia_gev.remote_device.node_map.TriggerMode.value = "On"
        self.ia_gev.remote_device.node_map.TriggerSource.value = "Software"

        ## Start image acquisition.
        self.ia_gev.start_acquisition()

    def close_blaze(self):
        ## Stop image acquisition.
        self.ia_blaze.stop_acquisition()
        self.ia_blaze.remote_device.node_map.TriggerMode.value = "Off"

        ## Disconnect from camera.
        self.ia_blaze.destroy()

    def close_2DCamera(self):
        ## Stop image acquisition.
        self.ia_gev.stop_acquisition()
        self.ia_gev.remote_device.node_map.TriggerMode.value = "Off"

        ## Disconnect from camera.
        self.ia_gev.destroy()

    def close_harvesters(self):
        ## Remove the CTI file and reset Harvester.
        self.h.reset()

    def get_image_blaze(self):
        with self.ia_blaze.fetch_buffer() as buffer:
            ## Warning: The buffer is only valid in the with statement and will be destroyed
            ## when you leave the scope.
            ## If you want to use the buffers outside of the with scope, you have to use np.copy()
            ## to make a deep copy of the image.

            ## Create an alias of the image components:
            pointcloud = buffer.payload.components[0]

            intensity = buffer.payload.components[1]

            confidence = buffer.payload.components[2]

            ## Reshape the range map into a 2D/3D array:
            ## "num_components_per_pixel" depends on the pixel format selected:
            ## "Coord3D_ABC32f" = 3
            ## "Coord3D_C16" = 1
            _3d = pointcloud.data.reshape(pointcloud.height, pointcloud.width,\
                                            int(pointcloud.num_components_per_pixel))

            ## Reshape the intensity image into a 2D array:
            _2d_intensity = intensity.data.reshape(intensity.height, intensity.width)

            ## Reshape the confidence image into a 2D array:
            _2d_confidence = confidence.data.reshape(confidence.height, confidence.width)

            ## Show the captured images as grayscale.
            cv2.imshow('range', _3d)

            cv2.imshow('_2d_intensity', _2d_intensity)

            cv2.imshow('_2d_confidence', _2d_confidence)

    def get_image_2DCamera(self):
        with self.ia_gev.fetch_buffer() as buffer:
            ## Warning: The buffer is only valid in the with statement and will be destroyed
            ## when you leave the scope.
            ## If you want to use the buffers outside of the with scope, you have to use np.copy()
            ## to make a deep copy of the image.

            ## Create an alias of the image component:
            image = buffer.payload.components[0]

            ## Reshape the image into a 2D array:
            _2d = image.data.reshape(image.height, image.width)

            ## Debayer the image to an RGB image.
            _2d_rgb = cv2.cvtColor(_2d, BAYER_FORMATS[image.data_format])

            ## Show the captured image as RGB.
            cv2.imshow('_2d_rgb', _2d_rgb)

    def run(self):
        ## Set up the cameras.
        self.setup_blaze()
        self.setup_2Dcamera()

        print('To exit, press ESC in one of the image windows')
        ## Grab the images.
        while True:
            ## To optimize bandwidth usage, the color camera is triggered first to 
            ## allow it to already transfer image data while the blaze camera is still internally
            ## processing the acquired raw data.
            self.ia_gev.remote_device.node_map.TriggerSoftware.execute()
            self.ia_blaze.remote_device.node_map.TriggerSoftware.execute()

            self.get_image_blaze()
            self.get_image_2DCamera()

            ## Break the endless loop by pressing ESC.
            k = cv2.waitKey(5) & 0xFF
            if k == 27:
                break

        ## Close the camera and release the producers.
        self.close_blaze()
        self.close_2DCamera()
        self.close_harvesters()

if __name__ == "__main__":
    ## Run the sample.
    Sample = ColorAndDepth()
    Sample.run()
